;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "fzf" "20260505.1105"
  "A front-end for fzf."
  '((emacs "24.4"))
  :url "https://github.com/bling/fzf.el"
  :commit "75060d81f4eea3f8088ac2a4a3a253c686b44d44"
  :revdesc "75060d81f4ee"
  :keywords '("fzf" "fuzzy" "search"))
